import com.maxeler.maxcompiler.v1.kernelcompiler.Kernel;
import com.maxeler.maxcompiler.v1.kernelcompiler.KernelParameters;
import com.maxeler.maxcompiler.v1.kernelcompiler.stdlib.KernelMath;
import com.maxeler.maxcompiler.v1.kernelcompiler.stdlib.core.Count;
import com.maxeler.maxcompiler.v1.kernelcompiler.types.base.HWVar;
import com.maxeler.maxcompiler.v1.kernelcompiler.types.composite.KArray;
import com.maxeler.maxcompiler.v1.kernelcompiler.types.composite.KArrayType;
import com.maxeler.maxcompiler.v1.kernelcompiler.stdlib.Bitops;


public class A4EnergyBoundKernel extends Kernel {

 
        private final int a,b,c,width;
        
        public KArrayType<HWVar> array_type = new KArrayType<HWVar>(hwInt(32), 4);
        public KArrayType<HWVar> array_type1 = new KArrayType<HWVar>(hwInt(32), 4);
        
        public A4EnergyBoundKernel(KernelParameters parameters) {
            super(parameters);
                        
            HWVar P1=io.scalarInput("P1", hwInt(32));
            HWVar P2=io.scalarInput("P2", hwInt(32));
            HWVar P3=io.scalarInput("P3", hwInt(32));
            HWVar P4=io.scalarInput("P4", hwInt(32));
            
            HWVar EMAX=io.scalarInput("EMAX", hwInt(32));
        
            /*
             If a non-dynamic energy-bound is used, we should use:
             HWVar E1=io.scalarinput("EM", hwInt(32));
             HWVar E2=E1;
         
             If a dynamic energy-bound is used, we should use:
             HWVar E1=io.input("E1", hwInt(32));
             HWVar E2=io.input("E2", hwInt(32));
             */
        
            Count.Params params1 = control.count.makeParams(width).withMax(2);
            
            Count.Counter count1=control.count.makeCounter(params1);
            
            Count.Params params2 = control.count.makeParams(width)
            .withEnable(count1.getWrap()).withMax(E1);
        
            Count.Counter count2 = control.count.makeCounter(params2);
            
            Count.Params params3 = control.count.makeParams(width)
            .withEnable(count2.getWrap()).withMax(E2);
            
            Count.Counter count3 = control.count.makeCounter(params3);
            
            Count.Params params4 = control.count.makeParams(width)
            .withEnable(count3.getWrap());
            
            Count.Counter startingstate = control.count.makeCounter(params4);
            
            HWVar BI=count1.getCount().eq(0)&count2.getCount().eq(0)&count3.getCount().eq(0);
            
            KArray<HWVar> n = io.input("n", array_type,BI);
            KArray<HWVar> l = io.input("l", array_type,BI);
            KArray<HWVar> j = io.input("j", array_type,BI);
            KArray<HWVar> m = io.input("m", array_type,BI);
            
            
            KArray<HWVar> out1 = array_type1.newInstance(this);
            KArray<HWVar> out2 = array_type1.newInstance(this);
            KArray<HWVar> out3 = array_type1.newInstance(this);
            KArray<HWVar> out4 = array_type1.newInstance(this);
            
            KArray<HWVar> ordering =array_type.newInstance(this);
            
            HWVar E = 2*n[0]+2*n[1]+2*n[2]+2*n[3]+l[0]+l[1]+l[2]+l[3];
            
            ordering[0]<==P1*n[0]+P2*l[0]+P3*j[0]+P4*m[0];
            ordering[1]<==P1*n[1]+P2*l[1]+P3*j[1]+P4*m[1];
            ordering[2]<==P1*n[2]+P2*l[2]+P3*j[2]+P4*m[2];
            ordering[3]<==P1*n[3]+P2*l[3]+P3*j[3]+P4*m[3];
            
            HWVar newj=count1.getCount().cast(hwInt(32));
            HWVar newn=count3.getCount().cast(hwInt(32));
            
            //Particle 1
            
            HWVar newl_1=2*count2.getCount().cast(hwInt(32))+KernelMath.modulo(l[0],2).cast(hwInt(32));
            
            HWVar ord_1=P1*newn+P2*newl_1+P3*newj+P4*m[0];
            
            HWVar col_1=ord_1.neq(ordering[1])&ord_1.neq(ordering[2])&ord_1.neq(ordering[3]);
            HWVar nE_1=EMAX-E+l[0]+2*n[0]-2*newn-newl_1;
            HWVar BO_1=col_1&(nE_1>-1)&(2*(newl_1+newj)>KernelMath.abs(m[0]));
            
            HWVar sign_1=(ord_1<ordering[1])?constant.var(hwInt(32),0):((ord_1<ordering[2])?constant.var(hwInt(32),1):((ord_1<ordering[3])?constant.var(hwInt(32),2):constant.var(hwInt(32),3)));
            
            HWVar p1_1=(ord_1<ordering[1])?ord_1:ordering[1];
            HWVar p2_1=(ord_1<ordering[2])?((ord_1<ordering[1])?ordering[1]:ord_1):ordering[1];
            HWVar p3_1=(ord_1<ordering[3])?((ord_1<ordering[2])?ordering[2]:ord_1):ordering[2];
            HWVar p4_1=(ord_1>ordering[3])?ord_1:ordering[3];
            
            HWVar key0_1=Bitops.concat(p1_1,p2_1,p3_1,p4_1);
            
            HWVar key1_1=key0_1^(key0_1>>a);
            HWVar key2_1=key1_1^(key1_1<<b);
            HWVar key3_1=key2_1^(key2_1>>c);
            
            HWVar key_1=key0_1.slice(0,32);
            
            HWVar addr1=key3_1.slice(0,32)^(key3_1.slice(32,32)^(key3_1.slice(64,32)^key3_1.slice(96,32)));
            
            out1[0]<==BO_1?addr1.cast(hwInt(32)):-1;
            out1[1]<==BO_1?key_1.cast(hwInt(32)):-1;
            out1[2]<==BO_1?sign_1:-1;
            out1[3]<==BO_1?startingstate.getCount().cast(hwInt(32)):-1;
 
            io.output("out1",out1,array_type1);
            
            
            //Particle 2
            
            HWVar newl_2=2*count2.getCount().cast(hwInt(32))+KernelMath.modulo(l[1],2).cast(hwInt(32));
            
            HWVar ord_2=P1*newn+P2*newl_2+P3*newj+P4*m[1];
        
            HWVar col_2=ord_2.neq(ordering[0])&ord_2.neq(ordering[1])&ord_2.neq(ordering[2])&ord_2.neq(ordering[3]);
            HWVar nE_2=EMAX-E+l[1]+2*n[1]-2*newn-newl_2;
            HWVar BO_2=col_2&(nE_2>-1)&(2*(newl_2+newj)>KernelMath.abs(m[1]));
        
            HWVar sign_2=(ord_2<ordering[0])?constant.var(hwInt(32),-1):((ord_2<ordering[2])?constant.var(hwInt(32),0):((ord_2<ordering[3])?constant.var(hwInt(32),1):constant.var(hwInt(32),2)));
        
            HWVar p1_2=(ord_2<ordering[0])?ord_2:ordering[0];
            HWVar p2_2=(ord_2<ordering[2])?((ord_2<ordering[0])?ordering[0]:ord_2):ordering[2];
            HWVar p3_2=(ord_2<ordering[3])?((ord_2<ordering[2])?ordering[2]:ord_2):ordering[3];
            HWVar p4_2=(ord_2>ordering[3])?ord_2:ordering[3];
            
            HWVar key0_2=Bitops.concat(p1_2,p2_2,p3_2,p4_2);
            
            HWVar key1_2=key0_2^(key0_2>>a);
            HWVar key2_2=key1_2^(key1_2<<b);
            HWVar key3_2=key2_2^(key2_2>>c);
        
            HWVar key_2=key0_2.slice(0,32);
            
            HWVar addr2=key3_2.slice(0,32)^(key3_2.slice(32,32)^(key3_2.slice(64,32)^key3_2.slice(96,32)));
            
            
            out2[0]<==BO_2?addr2.cast(hwInt(32)):-1;
            out2[1]<==BO_2?key_2.cast(hwInt(32)):-1;
            out2[2]<==BO_2?sign_2:-1;
            out2[3]<==BO_2?startingstate.getCount().cast(hwInt(32)):-1;
            
            io.output("out2",out2,array_type1);
            
            //Particle 3
            
            HWVar newl_3=2*count2.getCount().cast(hwInt(32))+KernelMath.modulo(l[2],2).cast(hwInt(32));
            
            HWVar ord_3=P1*newn+P2*newl_3+P3*newj+P4*m[2];
            
            HWVar col_3=ord_3.neq(ordering[0])&ord_3.neq(ordering[1])&ord_3.neq(ordering[2])&ord_3.neq(ordering[3]);
            HWVar nE_3=EMAX-E+l[2]+2*n[2]-2*newn-newl_3;
            HWVar BO_3=col_3&(nE_3>-1)&(2*(newl_3+newj)>KernelMath.abs(m[2]));
            
            HWVar sign_3=(ord_3<ordering[0])?constant.var(hwInt(32),-2):((ord_3<ordering[1])?constant.var(hwInt(32),-1):((ord_3<ordering[3])?constant.var(hwInt(32),0):constant.var(hwInt(32),1)));
            
            HWVar p1_3=(ord_3<ordering[0])?ord_3:ordering[0];
            HWVar p2_3=(ord_3<ordering[1])?((ord_3<ordering[0])?ordering[1]:ord_3):ordering[1];
            HWVar p3_3=(ord_3<ordering[3])?((ord_3<ordering[1])?ordering[1]:ord_3):ordering[3];
            HWVar p4_3=(ord_3>ordering[3])?ord_3:ordering[3];
            
            HWVar key0_3=Bitops.concat(p1_3,p2_3,p3_3,p4_3);
            
            HWVar key1_3=key0_3^(key0_3>>a);
            HWVar key2_3=key1_3^(key1_3<<b);
            HWVar key3_3=key2_3^(key2_3>>c);
            
            HWVar key_3=key0_3.slice(0,32);
            
            HWVar addr3=key3_3.slice(0,32)^(key3_3.slice(32,32)^(key3_3.slice(64,32)^key3_3.slice(96,32)));
            
            out3[0]<==BO_3?addr3.cast(hwInt(32)):-1;
            out3[1]<==BO_3?key_3.cast(hwInt(32)):-1;
            out3[2]<==BO_3?sign_3:-1;
            out3[3]<==BO_3?startingstate.getCount().cast(hwInt(32)):-1;
            
            io.output("out3",out3,array_type1);
        
            //Particle 4
            
            HWVar newl_4=2*count2.getCount().cast(hwInt(32))+KernelMath.modulo(l[3],2).cast(hwInt(32));
            
            HWVar ord_4=P1*newn+P2*newl_4+P3*newj+P4*m[3];
            
            HWVar col_4=ord_4.neq(ordering[0])&ord_4.neq(ordering[1])&ord_4.neq(ordering[2])&ord_4.neq(ordering[3]);
            HWVar nE_4=EMAX-E+l[3]+2*n[3]-2*newn-newl_4;
            HWVar BO_4=col_4&(nE_4>-1)&(2*(newl_4+newj)>KernelMath.abs(m[3]));
            
            HWVar sign_4=(ord_4<ordering[0])?constant.var(hwInt(32),-3):((ord_4<ordering[1])?constant.var(hwInt(32),-2):((ord_4<ordering[3])?constant.var(hwInt(32),-1):constant.var(hwInt(32),0)));
            
            HWVar p1_4=(ord_4<ordering[0])?ord_4:ordering[0];
            HWVar p2_4=(ord_4<ordering[1])?((ord_4<ordering[0])?ordering[1]:ord_4):ordering[1];
            HWVar p3_4=(ord_4<ordering[2])?((ord_4<ordering[1])?ordering[2]:ord_4):ordering[2];
            HWVar p4_4=(ord_4>ordering[2])?ord_4:ordering[2];
            
            HWVar key0_4=Bitops.concat(p1_4,p2_4,p3_4,p4_4);
            
            HWVar key1_4=key0_4^(key0_4>>a);
            HWVar key2_4=key1_4^(key1_4<<b);
            HWVar key3_4=key2_4^(key2_4>>c);
            
            HWVar key_4=key0_4.slice(0,32);
            
            HWVar addr4=key3_4.slice(0,32)^(key3_4.slice(32,32)^(key3_4.slice(64,32)^key3_4.slice(96,32)));
        
            out4[0]<==BO_4?addr4.cast(hwInt(32)):-1;
            out4[1]<==BO_4?key_4.cast(hwInt(32)):-1;
            out4[2]<==BO_4?sign_4:-1;
            out4[3]<==BO_4?startingstate.getCount().cast(hwInt(32)):-1;
        
            io.output("out4",out4,array_type1);
        }
}

