import com.maxeler.maxcompiler.v1.kernelcompiler.Kernel;
import com.maxeler.maxcompiler.v1.kernelcompiler.KernelParameters;
import com.maxeler.maxcompiler.v1.kernelcompiler.stdlib.KernelMath;
import com.maxeler.maxcompiler.v1.kernelcompiler.stdlib.core.Count;
import com.maxeler.maxcompiler.v1.kernelcompiler.types.base.HWVar;
import com.maxeler.maxcompiler.v1.kernelcompiler.types.composite.KArray;
import com.maxeler.maxcompiler.v1.kernelcompiler.types.composite.KArrayType;
import com.maxeler.maxcompiler.v1.kernelcompiler.stdlib.Bitops;


public class A2EnergyBoundKernel extends Kernel {

    private final int width;

    public KArrayType<HWVar> array_type = new KArrayType<HWVar>(hwInt(32), 2);
    public KArrayType<HWVar> array_type1 = new KArrayType<HWVar>(hwInt(32), 4);


    public A2EnergyBoundKernel(KernelParameters parameters) {
	super(parameters);
        
        
        HWVar P1=io.scalarInput("P1", hwInt(32));
        HWVar P2=io.scalarInput("P2", hwInt(32));
        HWVar P3=io.scalarInput("P3", hwInt(32));
        HWVar P4=io.scalarInput("P4", hwInt(32));
        
        HWVar EMAX=io.scalarInput("EMAX", hwInt(32));
        
        /*
         If a non-dynamic energy-bound is used, we should use:
         HWVar E1=io.scalarinput("EM", hwInt(32));
         HWVar E2=E1;
         
         If a dynamic energy-bound is used, we should use:
         HWVar E1=io.input("E1", hwInt(32));
         HWVar E2=io.input("E2", hwInt(32));
         */
        
        Count.Params params1 = control.count.makeParams(width).withMax(2);
        
        Count.Counter count1=control.count.makeCounter(params1);
        
        Count.Params params2 = control.count.makeParams(width)
	    .withEnable(count1.getWrap()).withMax(E1);
        
        
        Count.Counter count2 = control.count.makeCounter(params2);
        
        Count.Params params3 = control.count.makeParams(width)
	    .withEnable(count2.getWrap()).withMax(E2);
        
        Count.Counter count3 = control.count.makeCounter(params3);
        
        Count.Params params4 = control.count.makeParams(width)
	    .withEnable(count3.getWrap());
        
        Count.Counter startingstate = control.count.makeCounter(params4);
        
        HWVar BI=count1.getCount().eq(0)&count2.getCount().eq(0)&count3.getCount().eq(0);
        
        KArray<HWVar> n = io.input("n", array_type,BI);
        KArray<HWVar> l = io.input("l", array_type,BI);
        KArray<HWVar> j = io.input("j", array_type,BI);
        KArray<HWVar> m = io.input("m", array_type,BI);
        
        
        KArray<HWVar> out1 = array_type1.newInstance(this);
        KArray<HWVar> out2 = array_type1.newInstance(this);
        
        KArray<HWVar> ordering =array_type.newInstance(this);
        
        HWVar E = 2*n[0]+2*n[1]+l[0]+l[1];
        
        ordering[0]<==P1*n[0]+P2*l[0]+P3*j[0]+P4*m[0];
        ordering[1]<==P1*n[1]+P2*l[1]+P3*j[1]+P4*m[1];
        
        HWVar newj=count1.getCount().cast(hwInt(32));
        HWVar newn=count3.getCount().cast(hwInt(32));
        
        //Particle 1
        
        HWVar newl_1=2*count2.getCount().cast(hwInt(32))+KernelMath.modulo(l[0],2).cast(hwInt(32));
        
        HWVar ord_1=P1*newn+P2*newl_1+P3*newj+P4*m[0];
        
        HWVar col_1=ord_1.neq(ordering[1]);
        HWVar nE_1=EMAX-E+l[0]+2*n[0]-2*newn-newl_1;
        HWVar BO_1=col_1&(nE_1>-1)&(2*(newl_1+newj)>KernelMath.abs(m[0]));
                              
        HWVar sign_1=(ord_1>ordering[1])?constant.var(hwInt(32),1):constant.var(hwInt(32),0);
 
        out1[0]<==BO_1?ord_1:-1;
        out1[1]<==BO_1?ordering[0]:-1;
        out1[2]<==BO_1?sign_1:-1;
        out1[3]<==BO_1?startingstate.getCount().cast(hwInt(32)):-1;
  
        io.output("out1",out1,array_type1);
        
        //Particle 2
        
        HWVar newl_2=2*count2.getCount().cast(hwInt(32))+KernelMath.modulo(l[1],2).cast(hwInt(32));
        
        HWVar ord_2=P1*newn+P2*newl_2+P3*newj+P4*m[1];
        
        HWVar col_2=ord_2.neq(ordering[0])&ord_2.neq(ordering[1]);
        HWVar nE_2=EMAX-E+l[1]+2*n[1]-2*newn-newl_2;
        HWVar BO_2=col_2&(nE_2>-1)&(2*(newl_2+newj)>KernelMath.abs(m[1]));
        
        HWVar sign_2=ord_2>ordering[0]?constant.var(hwInt(32),0):constant.var(hwInt(32),-1);
        
        out2[0]<==BO_2?ord_2:-1;
        out2[1]<==BO_2?ordering[1]:-1;
        out2[2]<==BO_2?sign_2:-1;
        out2[3]<==BO_2?startingstate.getCount().cast(hwInt(32)):-1;
        
        io.output("out2",out2,array_type1);
    }
}

