import com.maxeler.maxcompiler.v1.kernelcompiler.Kernel;
import com.maxeler.maxcompiler.v1.kernelcompiler.KernelParameters;
import com.maxeler.maxcompiler.v1.kernelcompiler.stdlib.KernelMath;
import com.maxeler.maxcompiler.v1.kernelcompiler.stdlib.core.Count;
import com.maxeler.maxcompiler.v1.kernelcompiler.types.base.HWVar;
import com.maxeler.maxcompiler.v1.kernelcompiler.types.composite.KArray;
import com.maxeler.maxcompiler.v1.kernelcompiler.types.composite.KArrayType;
import com.maxeler.maxcompiler.v1.kernelcompiler.stdlib.Bitops;

public class A8EnergyBoundKernel extends Kernel {


    private final int width;

    public KArrayType<HWVar> array_type = new KArrayType<HWVar>(hwInt(32), 8);
    public KArrayType<HWVar> array_type1 = new KArrayType<HWVar>(hwInt(32), 4);


    public A8EnergyBoundKernel(KernelParameters parameters) {
	super(parameters);
        
        HWVar P1=io.scalarInput("P1", hwInt(32));
        HWVar P2=io.scalarInput("P2", hwInt(32));
        HWVar P3=io.scalarInput("P3", hwInt(32));
        HWVar P4=io.scalarInput("P4", hwInt(32));
        
        /*
         If a non-dynamic energy-bound is used, we should use:
         HWVar E1=io.scalarinput("EM", hwInt(32));
         HWVar E2=E1;
         
         If a dynamic energy-bound is used, we should use:
         HWVar E1=io.input("E1", hwInt(32));
         HWVar E2=io.input("E2", hwInt(32));
         */
        
        HWVar EMAX=io.scalarInput("EMAX", hwInt(32));
        
        Count.Params params1 = control.count.makeParams(width).withMax(2);
        
        Count.Counter count1=control.count.makeCounter(params1);
        
        Count.Params params2 = control.count.makeParams(width)
	    .withEnable(count1.getWrap()).withMax(E1);
        
        
        Count.Counter count2 = control.count.makeCounter(params2);
        
        Count.Params params3 = control.count.makeParams(width)
	    .withEnable(count2.getWrap()).withMax(E2);
        
        Count.Counter count3 = control.count.makeCounter(params3);
        
        Count.Params params4 = control.count.makeParams(width)
	    .withEnable(count3.getWrap());
        
        Count.Counter startingstate = control.count.makeCounter(params4);
        
        HWVar BI=count1.getCount().eq(0)&count2.getCount().eq(0)&count3.getCount().eq(0);
        
        KArray<HWVar> n = io.input("n", array_type,BI);
        KArray<HWVar> l = io.input("l", array_type,BI);
        KArray<HWVar> j = io.input("j", array_type,BI);
        KArray<HWVar> m = io.input("m", array_type,BI);
        
        
        KArray<HWVar> out1 = array_type1.newInstance(this);
        KArray<HWVar> out2 = array_type1.newInstance(this);
        KArray<HWVar> out3 = array_type1.newInstance(this);
        KArray<HWVar> out4 = array_type1.newInstance(this);
        KArray<HWVar> out5 = array_type1.newInstance(this);
        KArray<HWVar> out6 = array_type1.newInstance(this);
        KArray<HWVar> out7 = array_type1.newInstance(this);
        KArray<HWVar> out8 = array_type1.newInstance(this);
        
        KArray<HWVar> ordering =array_type.newInstance(this);
        
        HWVar E = 2*n[0]+2*n[1]+2*n[2]+2*n[3]+2*n[4]+2*n[5]+2*n[6]+2*n[7]+l[0]+l[1]+l[2]+l[3]+l[4]+l[5]+l[6]+l[7];
        
        ordering[0]<==P1*n[0]+P2*l[0]+P3*j[0]+P4*m[0];
        ordering[1]<==P1*n[1]+P2*l[1]+P3*j[1]+P4*m[1];
        ordering[2]<==P1*n[2]+P2*l[2]+P3*j[2]+P4*m[2];
        ordering[3]<==P1*n[3]+P2*l[3]+P3*j[3]+P4*m[3];
        ordering[4]<==P1*n[4]+P2*l[4]+P3*j[4]+P4*m[4];
        ordering[5]<==P1*n[5]+P2*l[5]+P3*j[5]+P4*m[5];
        ordering[6]<==P1*n[6]+P2*l[6]+P3*j[6]+P4*m[6];
        ordering[7]<==P1*n[7]+P2*l[7]+P3*j[7]+P4*m[7];
        
        HWVar newj=count1.getCount().cast(hwInt(32));
        HWVar newn=count3.getCount().cast(hwInt(32));
        
        //Particle 1
        
        HWVar newl_1=2*count2.getCount().cast(hwInt(32))+KernelMath.modulo(l[0],2).cast(hwInt(32));
        
        HWVar ord_1=P1*newn+P2*newl_1+P3*newj+P4*m[0];
        
        HWVar col_1=ord_1.neq(ordering[1])&ord_1.neq(ordering[2])&ord_1.neq(ordering[3])&ord_1.neq(ordering[4])&ord_1.neq(ordering[5])&ord_1.neq(ordering[6])&ord_1.neq(ordering[7]);
        HWVar nE_1=EMAX-E+l[0]+2*n[0]-2*newn-newl_1;
        HWVar BO_1=col_1&(nE_1>-1)&(2*(newl_1+newj)>KernelMath.abs(m[0]));
        
        HWVar sign_1=(ord_1<ordering[1])?constant.var(hwInt(32),0):((ord_1<ordering[2])?constant.var(hwInt(32),1):((ord_1<ordering[3])?constant.var(hwInt(32),2):((ord_1<ordering[4])?constant.var(hwInt(32),3):((ord_1<ordering[5])?constant.var(hwInt(32),4):((ord_1<ordering[6])?constant.var(hwInt(32),5):((ord_1<ordering[7])?constant.var(hwInt(32),6):constant.var(hwInt(32),7)))))));
        
        out1[0]<==BO_1?ord_1:-1;
        out1[1]<==BO_1?ordering[0]:-1;
        out1[2]<==BO_1?sign_1:-1;
        out1[3]<==BO_1?startingstate.getCount().cast(hwInt(32)):-1;
        
        io.output("out1",out1,array_type1);
        
        //Particle 2
        
        HWVar newl_2=2*count2.getCount().cast(hwInt(32))+KernelMath.modulo(l[1],2).cast(hwInt(32));
        
        HWVar ord_2=P1*newn+P2*newl_2+P3*newj+P4*m[1];
        
        HWVar col_2=ord_2.neq(ordering[0])&ord_2.neq(ordering[1])&ord_2.neq(ordering[2])&ord_2.neq(ordering[3])&ord_2.neq(ordering[4])&ord_2.neq(ordering[5])&ord_2.neq(ordering[6])&ord_2.neq(ordering[7]);
        HWVar nE_2=EMAX-E+l[1]+2*n[1]-2*newn-newl_2;
        HWVar BO_2=col_2&(nE_2>-1)&(2*(newl_2+newj)>KernelMath.abs(m[1]));
        
        HWVar sign_2=(ord_2<ordering[0])?constant.var(hwInt(32),-1):((ord_2<ordering[2])?constant.var(hwInt(32),0):((ord_2<ordering[3])?constant.var(hwInt(32),1):((ord_2<ordering[4])?constant.var(hwInt(32),2):((ord_2<ordering[5])?constant.var(hwInt(32),3):((ord_2<ordering[6])?constant.var(hwInt(32),4):((ord_2<ordering[7])?constant.var(hwInt(32),5):constant.var(hwInt(32),6)))))));
        
        out2[0]<==BO_2?ord_2:-1;
        out2[1]<==BO_2?ordering[1]:-1;
        out2[2]<==BO_2?sign_2:-1;
        out2[3]<==BO_2?startingstate.getCount().cast(hwInt(32)):-1;
        
        io.output("out2",out2,array_type1);
        
        //Particle 3
        
        HWVar newl_3=2*count2.getCount().cast(hwInt(32))+KernelMath.modulo(l[2],2).cast(hwInt(32));
    
        HWVar ord_3=P1*newn+P2*newl_3+P3*newj+P4*m[2];

        HWVar col_3=ord_3.neq(ordering[0])&ord_3.neq(ordering[1])&ord_3.neq(ordering[2])&ord_3.neq(ordering[3])&ord_3.neq(ordering[4])&ord_3.neq(ordering[5])&ord_3.neq(ordering[6])&ord_3.neq(ordering[7]);
        HWVar nE_3=EMAX-E+l[2]+2*n[2]-2*newn-newl_3;
        HWVar BO_3=col_3&(nE_3>-1)&(2*(newl_3+newj)>KernelMath.abs(m[2]));
        
        HWVar sign_3=(ord_3<ordering[0])?constant.var(hwInt(32),-2):((ord_3<ordering[1])?constant.var(hwInt(32),-1):((ord_3<ordering[3])?constant.var(hwInt(32),0):((ord_3<ordering[4])?constant.var(hwInt(32),1):((ord_3<ordering[5])?constant.var(hwInt(32),2):((ord_3<ordering[6])?constant.var(hwInt(32),3):((ord_3<ordering[7])?constant.var(hwInt(32),4):constant.var(hwInt(32),5)))))));
        
        out3[0]<==BO_3?ord_3:-1;
        out3[1]<==BO_3?ordering[2]:-1;
        out3[2]<==BO_3?sign_3:-1;
        out3[3]<==BO_3?startingstate.getCount().cast(hwInt(32)):-1;
        
        io.output("out3",out3,array_type1);
        
        //Particle 4
        
        HWVar newl_4=2*count2.getCount().cast(hwInt(32))+KernelMath.modulo(l[3],2).cast(hwInt(32));
        
        HWVar ord_4=P1*newn+P2*newl_4+P3*newj+P4*m[3];
        
        HWVar col_4=ord_4.neq(ordering[0])&ord_4.neq(ordering[1])&ord_4.neq(ordering[2])&ord_4.neq(ordering[3])&ord_4.neq(ordering[4])&ord_4.neq(ordering[5])&ord_4.neq(ordering[6])&ord_4.neq(ordering[7]);
        HWVar nE_4=EMAX-E+l[3]+2*n[3]-2*newn-newl_4;
        HWVar BO_4=col_4&(nE_4>-1)&(2*(newl_4+newj)>KernelMath.abs(m[3]));
        
        HWVar sign_4=(ord_4<ordering[0])?constant.var(hwInt(32),-3):((ord_4<ordering[1])?constant.var(hwInt(32),-2):((ord_4<ordering[2])?constant.var(hwInt(32),-1):((ord_4<ordering[4])?constant.var(hwInt(32),0):((ord_4<ordering[5])?constant.var(hwInt(32),1):((ord_4<ordering[6])?constant.var(hwInt(32),2):((ord_4<ordering[7])?constant.var(hwInt(32),3):constant.var(hwInt(32),4)))))));
        
        out4[0]<==BO_4?ord_4:-1;
        out4[1]<==BO_4?ordering[3]:-1;
        out4[2]<==BO_4?sign_4:-1;
        out4[3]<==BO_4?startingstate.getCount().cast(hwInt(32)):-1;
        
        io.output("out4",out4,array_type1);
        
        //Particle 5
        
        HWVar newl_5=2*count2.getCount().cast(hwInt(32))+KernelMath.modulo(l[4],2).cast(hwInt(32));
        
        HWVar ord_5=P1*newn+P2*newl_5+P3*newj+P4*m[4];
        
        HWVar col_5=ord_5.neq(ordering[0])&ord_5.neq(ordering[1])&ord_5.neq(ordering[2])&ord_5.neq(ordering[3])&ord_5.neq(ordering[4])&ord_5.neq(ordering[5])&ord_5.neq(ordering[6])&ord_5.neq(ordering[7]);
        HWVar nE_5=EMAX-E+l[4]+2*n[4]-2*newn-newl_5;
        HWVar BO_5=col_5&(nE_5>-1)&(2*(newl_5+newj)>KernelMath.abs(m[4]));
        HWVar sign_5=(ord_5<ordering[0])?constant.var(hwInt(32),-4):((ord_5<ordering[1])?constant.var(hwInt(32),-3):((ord_5<ordering[2])?constant.var(hwInt(32),-2):((ord_5<ordering[3])?constant.var(hwInt(32),-1):((ord_5<ordering[5])?constant.var(hwInt(32),0):((ord_5<ordering[6])?constant.var(hwInt(32),1):((ord_5<ordering[7])?constant.var(hwInt(32),2):constant.var(hwInt(32),3)))))));
        
        out5[0]<==BO_5?ord_5:-1;
        out5[1]<==BO_5?ordering[4]:-1;
        out5[2]<==BO_5?sign_5:-1;
        out5[3]<==BO_5?startingstate.getCount().cast(hwInt(32)):-1;
        
        io.output("out5",out5,array_type1);
        
        //Particle 6
        
        HWVar newl_6=2*count2.getCount().cast(hwInt(32))+KernelMath.modulo(l[5],2).cast(hwInt(32));
        
        HWVar ord_6=P1*newn+P2*newl_6+P3*newj+P4*m[5];
        
        HWVar col_6=ord_6.neq(ordering[0])&ord_6.neq(ordering[1])&ord_6.neq(ordering[2])&ord_6.neq(ordering[3])&ord_6.neq(ordering[4])&ord_6.neq(ordering[5])&ord_6.neq(ordering[6])&ord_6.neq(ordering[7]);
        HWVar nE_6=EMAX-E+l[5]+2*n[5]-2*newn-newl_6;
        HWVar BO_6=col_6&(nE_6>-1)&(2*(newl_6+newj)>KernelMath.abs(m[5]));
        
        HWVar sign_6=(ord_6<ordering[0])?constant.var(hwInt(32),-5):((ord_6<ordering[1])?constant.var(hwInt(32),-4):((ord_6<ordering[2])?constant.var(hwInt(32),-3):((ord_6<ordering[3])?constant.var(hwInt(32),-2):((ord_6<ordering[4])?constant.var(hwInt(32),-1):((ord_6<ordering[6])?constant.var(hwInt(32),0):((ord_6<ordering[7])?constant.var(hwInt(32),1):constant.var(hwInt(32),2)))))));
        
        out6[0]<==BO_6?ord_6:-1;
        out6[1]<==BO_6?ordering[5]:-1;
        out6[2]<==BO_6?sign_6:-1;
        out6[3]<==BO_6?startingstate.getCount().cast(hwInt(32)):-1;
        
        io.output("out6",out6,array_type1);
        
        //Partikel 7
        
        HWVar newl_7=2*count2.getCount().cast(hwInt(32))+KernelMath.modulo(l[6],2).cast(hwInt(32));
        
        HWVar ord_7=P1*newn+P2*newl_7+P3*newj+P4*m[6];
        
        HWVar col_7=ord_7.neq(ordering[0])&ord_7.neq(ordering[1])&ord_7.neq(ordering[2])&ord_7.neq(ordering[3])&ord_7.neq(ordering[4])&ord_7.neq(ordering[5])&ord_7.neq(ordering[6])&ord_7.neq(ordering[7]);
        HWVar nE_7=EMAX-E+l[6]+2*n[6]-2*newn-newl_7;
        HWVar BO_7=col_7&(nE_7>-1)&(2*(newl_7+newj)>KernelMath.abs(m[6]));
        
        HWVar sign_7=(ord_7<ordering[0])?constant.var(hwInt(32),-6):((ord_7<ordering[1])?constant.var(hwInt(32),-5):((ord_7<ordering[2])?constant.var(hwInt(32),-4):((ord_7<ordering[3])?constant.var(hwInt(32),-3):((ord_7<ordering[4])?constant.var(hwInt(32),-2):((ord_7<ordering[5])?constant.var(hwInt(32),-1):((ord_7<ordering[7])?constant.var(hwInt(32),0):constant.var(hwInt(32),1)))))));
 
        out7[0]<==BO_7?ord_7:-1;
        out7[1]<==BO_7?ordering[6]:-1;
        out7[2]<==BO_7?sign_7:-1;
        out7[3]<==BO_7?startingstate.getCount().cast(hwInt(32)):-1;
        
        io.output("out7",out7,array_type1);
        
        //Particle 8
        
        HWVar newl_8=2*count2.getCount().cast(hwInt(32))+KernelMath.modulo(l[7],2).cast(hwInt(32));
        
        HWVar ord_8=P1*newn+P2*newl_8+P3*newj+P4*m[7];
        
        HWVar col_8=ord_8.neq(ordering[0])&ord_8.neq(ordering[1])&ord_8.neq(ordering[2])&ord_8.neq(ordering[3])&ord_8.neq(ordering[4])&ord_8.neq(ordering[5])&ord_8.neq(ordering[6])&ord_8.neq(ordering[7]);
        HWVar nE_8=EMAX-E+l[7]+2*n[7]-2*newn-newl_8;
        HWVar BO_8=col_8&(nE_8>-1)&(2*(newl_8+newj)>KernelMath.abs(m[7]));
        
        HWVar sign_8=(ord_8<ordering[0])?constant.var(hwInt(32),-7):((ord_8<ordering[1])?constant.var(hwInt(32),-6):((ord_8<ordering[2])?constant.var(hwInt(32),-5):((ord_8<ordering[3])?constant.var(hwInt(32),-4):((ord_8<ordering[4])?constant.var(hwInt(32),-3):((ord_8<ordering[5])?constant.var(hwInt(32),-2):((ord_8<ordering[6])?constant.var(hwInt(32),-1):constant.var(hwInt(32),0)))))));
        
        out8[0]<==BO_8?ord_8:-1;
        out8[1]<==BO_8?ordering[7]:-1;
        out8[2]<==BO_8?sign_8:-1;
        out8[3]<==BO_8?startingstate.getCount().cast(hwInt(32)):-1;
        
        io.output("out8",out8,array_type1);
    
    }
}

